package checkmk

import (
	"context"
	"encoding/base64"
	"fmt"
	"log"
	"sync"

	"github.com/it-novum/nsta/gearman"
	jsoniter "github.com/json-iterator/go"
)

var json = jsoniter.ConfigCompatibleWithStandardLibrary

// CommandBackend should implement the execution of commands on the remote machine
type CommandBackend interface {
	CmkRunCommand(ctx context.Context, hostUUID string, args []string, env map[string]string) (string, error)
	CmkWriteFile(ctx context.Context, hostUUID, path string, content []byte) error
}

// Checkmk manages the execution of checkmk commands
type Checkmk struct {
	mtx           sync.Mutex
	SatelliteName string
}

func (c *Checkmk) healthScan(cmd CommandBackend, msg *gearman.CheckmkMessage, response *gearman.CheckmkResponse) {
	var err error
	log.Println(&Error{Name: c.SatelliteName, HostUUID: msg.Data.Hostuuid, Message: "performing checkmk health scan"})

	if msg.Data.File != "" && msg.Data.FilePath != "" {
		c.writeFile(cmd, msg, response)
		if response.Error != "" {
			return
		}
	}

	env := map[string]string{
		"PYTHONPATH": "/opt/openitc/check_mk/lib/python",
		"OMD_ROOT":   "/opt/openitc/check_mk",
		"OMD_SITE":   "1",
	}

	log.Println(&Error{Name: c.SatelliteName, HostUUID: msg.Data.Hostuuid, Message: "clear content of autochecks directory"})
	if _, err = cmd.CmkRunCommand(nil, msg.Data.Hostuuid, []string{"/bin/rm", "-rf", "/opt/openitc/check_mk/var/check_mk/autochecks/"}, env); err != nil {
		response.Error = fmt.Sprint("could not clear the contents of the autochecks folder: ", err)
		log.Println(err)
		return
	}

	log.Println(&Error{Name: c.SatelliteName, HostUUID: msg.Data.Hostuuid, Message: "create autochecks directory"})
	if _, err = cmd.CmkRunCommand(nil, msg.Data.Hostuuid, []string{"/bin/mkdir", "-p", "/opt/openitc/check_mk/var/check_mk/autochecks"}, env); err != nil {
		response.Error = fmt.Sprint("could not create autochecks folder: ", err)
		log.Println(err)
		return
	}

	log.Println(&Error{Name: c.SatelliteName, HostUUID: msg.Data.Hostuuid, Message: "run the inventory command"})
	if _, err = cmd.CmkRunCommand(nil, msg.Data.Hostuuid, []string{"/opt/openitc/check_mk/bin/check_mk", "-II", msg.Data.Hostuuid}, env); err != nil {
		response.Error = fmt.Sprint("could not run checkmk inventory: ", err)
		log.Println(err)
		return
	}

	log.Println(&Error{Name: c.SatelliteName, HostUUID: msg.Data.Hostuuid, Message: "receive inventory scan result"})
	if response.DumpHostResult, err = cmd.CmkRunCommand(nil, msg.Data.Hostuuid, []string{"/opt/openitc/check_mk/bin/check_mk", "-D", msg.Data.Hostuuid}, env); err != nil {
		response.Error = fmt.Sprint("could not dump host result: ", err)
		log.Println(err)
		return
	}

	log.Println(&Error{Name: c.SatelliteName, HostUUID: msg.Data.Hostuuid, Message: "receive check types"})
	if response.CheckTypesResult, err = cmd.CmkRunCommand(nil, msg.Data.Hostuuid, []string{"/opt/openitc/check_mk/bin/check_mk", "-L"}, env); err != nil {
		response.Error = fmt.Sprint("could not dump check types: ", err)
		log.Println(err)
		return
	}

	log.Println(&Error{Name: c.SatelliteName, HostUUID: msg.Data.Hostuuid, Message: "successful inventory scan"})
	return
}

func (c *Checkmk) processScan(cmd CommandBackend, msg *gearman.CheckmkMessage, response *gearman.CheckmkResponse) {
	var err error
	log.Println(&Error{Name: c.SatelliteName, HostUUID: msg.Data.Hostuuid, Message: "performing checkmk process scan"})

	if msg.Data.File != "" && msg.Data.FilePath != "" {
		c.writeFile(cmd, msg, response)
		if response.Error != "" {
			return
		}
	}

	env := map[string]string{
		"PYTHONPATH": "/opt/openitc/check_mk/lib/python",
		"OMD_ROOT":   "/opt/openitc/check_mk",
		"OMD_SITE":   "1",
	}
	if response.RawInfoResult, err = cmd.CmkRunCommand(nil, msg.Data.Hostuuid, []string{"/opt/openitc/check_mk/bin/check_mk", "-d", msg.Data.Hostuuid}, env); err != nil {
		response.Error = fmt.Sprint("could not dump process info: ", err)
		log.Println(err)
		return
	}
	log.Println(&Error{Name: c.SatelliteName, HostUUID: msg.Data.Hostuuid, Message: "successful process scan"})
	return
}

func (c *Checkmk) writeFile(cmd CommandBackend, msg *gearman.CheckmkMessage, response *gearman.CheckmkResponse) {
	log.Println(&Error{Name: c.SatelliteName, HostUUID: msg.Data.Hostuuid, Message: fmt.Sprint("writing file to: ", msg.Data.FilePath)})

	fd, err := base64.StdEncoding.DecodeString(msg.Data.File)
	if err != nil {
		e := &Error{Name: c.SatelliteName, HostUUID: msg.Data.Hostuuid, Message: "could not decode file", Err: err}
		log.Println(e)
		response.Error = e.Error()
		return
	}

	err = cmd.CmkWriteFile(nil, msg.Data.Hostuuid, msg.Data.FilePath, fd)
	if err != nil {
		log.Println(err)
		response.Error = err.Error()
	} else {
		log.Println(&Error{Name: c.SatelliteName, HostUUID: msg.Data.Hostuuid, Message: "write file successful"})
	}
}

// ProcessMessage parses a checkmk message and runs the required commands on the satellite
func (c *Checkmk) ProcessMessage(cmd CommandBackend, msg *gearman.Message) *gearman.CheckmkResponse {
	c.mtx.Lock()
	defer c.mtx.Unlock()

	m := &gearman.CheckmkMessage{}
	err := json.Unmarshal([]byte(msg.String), m)
	response := &gearman.CheckmkResponse{
		SatelliteID: m.SatelliteID,
		Command:     m.Command,
		ScanID:      m.Data.ScanID,
	}
	if err != nil {
		log.Println(&Error{Name: c.SatelliteName, Message: "could not parse message", Err: err})
		response.Error = fmt.Sprint("could not parse message for checkmk:", err)
		return response
	}

	switch m.Data.Task {
	case "health-scan":
		c.healthScan(cmd, m, response)
	case "snmp-scan":
		c.healthScan(cmd, m, response)
	case "process-scan":
		c.processScan(cmd, m, response)
	case "write-file":
		c.writeFile(cmd, m, response)
	default:
		log.Println(&Error{Name: c.SatelliteName, Message: fmt.Sprint("unknown task: ", m.Data.Task)})
		response.Error = fmt.Sprint("unknown task for checkmk: ", m.Data.Task)
	}

	return response
}
