package gearman

import (
	"context"
	"testing"
	"time"
)

func TestClientMain(t *testing.T) {
	msgReceive := make(chan *Message, 200)
	defer close(msgReceive)

	main := Client{
		ClientType:     Main,
		ClientName:     "test-main",
		Address:        "127.0.0.1:4730",
		MessageReceive: msgReceive,
	}

	main.Start(context.Background())
	defer main.Stop()

	timeout := time.NewTimer(time.Second * 10)
	defer timeout.Stop()

	for !main.IsRunning() {
		select {
		case <-timeout.C:
			t.Fatal("Timeout during startup")
		default:
			time.Sleep(time.Millisecond * 50)
		}
	}
}

func TestClientSatellite(t *testing.T) {
	msgReceive := make(chan *Message, 200)
	defer close(msgReceive)

	sat := Client{
		ClientType:     Satellite,
		ClientName:     "test-sat",
		Address:        "127.0.0.1:4730",
		MessageReceive: msgReceive,
	}
	sat.Start(context.Background())
	defer sat.Stop()

	timeout := time.NewTimer(time.Second * 10)
	defer timeout.Stop()

	for !sat.IsRunning() {
		select {
		case <-timeout.C:
			t.Fatal("Timeout during startup")
		default:
			time.Sleep(time.Millisecond * 50)
		}
	}
}

func TestCancel(t *testing.T) {
	msgReceive := make(chan *Message, 200)
	defer close(msgReceive)

	ctx, cancel := context.WithCancel(context.Background())

	main := Client{
		ClientType:     Main,
		ClientName:     "test-main",
		Address:        "127.0.0.1:4730",
		MessageReceive: msgReceive,
	}
	main.Start(ctx)

	timeout := time.NewTimer(time.Second * 10)
	defer timeout.Stop()

	for !main.IsRunning() {
		select {
		case <-timeout.C:
			t.Fatal("Timeout during startup")
		default:
			time.Sleep(time.Millisecond * 50)
		}
	}

	cancel()

	for main.IsRunning() {
		select {
		case <-timeout.C:
			t.Fatal("Timeout during shutdown")
		default:
			time.Sleep(time.Millisecond * 50)
		}
	}
}

func TestFull(t *testing.T) {
	msgReceive := make(chan *Message, 200)
	defer close(msgReceive)

	main := Client{
		ClientType:     Main,
		ClientName:     "test-main",
		Address:        "127.0.0.1:4730",
		MessageReceive: msgReceive,
	}

	main.Start(context.Background())
	defer main.Stop()

	timeout := time.NewTimer(time.Second * 10)
	defer timeout.Stop()

	for !main.IsRunning() {
		select {
		case <-timeout.C:
			t.Fatal("Timeout during startup")
		default:
			time.Sleep(time.Millisecond * 50)
		}
	}

	main.SendMessage(&Message{
		Type:   MessageCustomInput,
		String: "Test",
	})

	if err := main.sendRawMessage(MessageCommandSatTX.WorkerQueue(), "Test"); err != nil {
		t.Fatal(err)
	}

	select {
	case <-timeout.C:
		t.Fatal("Timeout during message receive")
	case msg := <-msgReceive:
		if msg.Type != MessageCommandSatTX {
			t.Fatal("unxpected message type: ", msg.Type)
		}
		if msg.String != "Test" {
			t.Fatal("unxpected message text: ", msg.String)
		}
	}
}

func TestInvalidAddress(t *testing.T) {
	msgReceive := make(chan *Message, 200)
	defer close(msgReceive)

	main := Client{
		ClientType:     Main,
		ClientName:     "test-main",
		Address:        "127.0.0.1:1",
		MessageReceive: msgReceive,
	}

	main.Start(context.Background())
	defer main.Stop()

	timeout := time.NewTimer(time.Second * 1)
	defer timeout.Stop()

	for !main.IsRunning() {
		select {
		case <-timeout.C:
			// success
			return
		default:
			time.Sleep(time.Millisecond * 50)
		}
	}
	t.Fatal("connection was successful")
}
