package gearman

import (
	"container/list"
	"sync"
)

// SyncMessageFIFO provides a synced FIFO list for gearman messages
type SyncMessageFIFO struct {
	noCopy noCopy

	mtx      sync.Mutex
	messages list.List

	input      chan *Message
	inputBack  chan []*Message
	inputFront chan []*Message
	notify     chan struct{}

	closer chan struct{}
	wg     sync.WaitGroup
}

func (f *SyncMessageFIFO) run() {
	defer func() {
		close(f.input)
		close(f.inputFront)
		close(f.inputBack)
		close(f.notify)
		close(f.closer)
		f.wg.Done()
	}()
	for {
		f.mtx.Lock()
		new := false
		if f.messages.Len() > 0 {
			new = true
		}
		f.mtx.Unlock()
		if new {
			select {
			case <-f.closer:
				return
			case f.notify <- struct{}{}:
			case msg := <-f.input:
				f.mtx.Lock()
				f.messages.PushBack(msg)
				f.mtx.Unlock()
			case msgs := <-f.inputFront:
				f.mtx.Lock()
				for _, msg := range msgs {
					f.messages.PushFront(msg)
				}
				f.mtx.Unlock()
			case msgs := <-f.inputBack:
				f.mtx.Lock()
				for _, msg := range msgs {
					f.messages.PushBack(msg)
				}
				f.mtx.Unlock()
			}
		} else {
			select {
			case <-f.closer:
				return
			case msg := <-f.input:
				f.mtx.Lock()
				f.messages.PushBack(msg)
				f.mtx.Unlock()
			case msgs := <-f.inputFront:
				f.mtx.Lock()
				for _, msg := range msgs {
					f.messages.PushFront(msg)
				}
				f.mtx.Unlock()
			case msgs := <-f.inputBack:
				f.mtx.Lock()
				for _, msg := range msgs {
					f.messages.PushBack(msg)
				}
				f.mtx.Unlock()
			}
		}
	}
}

// Start creates a background process for io i/o
func (f *SyncMessageFIFO) Start() {
	f.input = make(chan *Message)
	f.inputFront = make(chan []*Message)
	f.inputBack = make(chan []*Message)
	f.notify = make(chan struct{})
	f.closer = make(chan struct{})
	f.wg.Add(1)
	go f.run()
}

// Close sends a signal to the gopher to stop
func (f *SyncMessageFIFO) Close() {
	f.closer <- struct{}{}
	f.wg.Wait()
}

// Notify handles a notification if a new message is available in the buffer
func (f *SyncMessageFIFO) Notify() chan struct{} {
	return f.notify
}

// Push adds a message to the FIFO
func (f *SyncMessageFIFO) Push(msg *Message) {
	f.input <- msg
}

// PushAll adds a slice of messages to the FIFO
func (f *SyncMessageFIFO) PushAll(msgs []*Message, front bool) {
	if front {
		f.inputFront <- msgs
	} else {
		f.inputBack <- msgs
	}
}

// Get returns one elemnt from the FIFO
func (f *SyncMessageFIFO) Get() *Message {
	f.mtx.Lock()
	defer f.mtx.Unlock()

	el := f.messages.Front()
	if el == nil {
		return nil
	}
	msg := el.Value.(*Message)

	f.messages.Remove(el)
	return msg
}

// GetAll returns up to max messages from the FIFO
func (f *SyncMessageFIFO) GetAll(max int) []*Message {
	f.mtx.Lock()
	defer f.mtx.Unlock()

	lenMsg := f.messages.Len()
	if lenMsg > max {
		lenMsg = max
	}
	msgs := make([]*Message, lenMsg)

	for i := 0; i < lenMsg; i++ {
		el := f.messages.Front()
		if el == nil {
			msgs = msgs[:i]
			break
		}
		msgs[i] = f.messages.Remove(el).(*Message)
	}

	return msgs
}

// Direct returns the channel were you can push messages directly into the buffer
func (f *SyncMessageFIFO) Direct() chan<- *Message {
	return f.input
}
