package oitcdb

import (
	"fmt"
	"sync"
	"time"

	"github.com/jinzhu/gorm"
	// we always use mysql and have to explicit include it for dependency management
	_ "github.com/jinzhu/gorm/dialects/mysql"
	"gopkg.in/ini.v1"
)

// Satellite is the model for the oitc satellites table
type Satellite struct {
	ID                int `gorm:"PRIMARY_KEY"`
	Name              string
	Address           string
	ContainerID       int
	SyncInstance      int
	SyncMethod        string
	Timezone          string
	Created           time.Time
	Modified          time.Time
	Login             string
	Port              int
	PrivateKeyPath    string
	URL               string
	RemotePort        int
	APIKey            string
	Interval          int
	Timeout           int
	UseProxy          int
	ProxyURL          string
	VerifyCertificate bool
	UseTimesync       bool
	NSTASyncInstance  bool `gorm:"Column:nsta_sync_instance"`
}

// Compare all nsta important fields and return true if equal
func (s *Satellite) Compare(other *Satellite) bool {
	if s.ID == other.ID && s.Name == other.Name && s.Address == other.Address && s.NSTASyncInstance == other.NSTASyncInstance &&
		s.SyncMethod == other.SyncMethod && s.Login == other.Login && s.Port == other.Port &&
		s.PrivateKeyPath == other.PrivateKeyPath && s.URL == other.URL && s.RemotePort == other.RemotePort &&
		s.APIKey == other.APIKey && s.Interval == other.Interval && s.Timeout == other.Timeout &&
		s.UseProxy == other.UseProxy && s.ProxyURL == other.ProxyURL && s.VerifyCertificate == other.VerifyCertificate &&
		s.UseTimesync == other.UseTimesync {
		return true
	}
	return false
}

func NewSatelliteMap(satellites []*Satellite) map[int]*Satellite {
	s := make(map[int]*Satellite)
	for _, dbSat := range satellites {
		s[dbSat.ID] = dbSat
	}
	return s
}

type SatelliteStatusID int

const (
	SatelliteStatusNil    SatelliteStatusID = 0
	SatelliteStatusOK     SatelliteStatusID = 1
	SatelliteStatusFailed SatelliteStatusID = 2
)

// SatelliteStatus is the model for the oitc satellite_status table
type SatelliteStatus struct {
	db            *gorm.DB
	ID            int `gorm:"PRIMARY_KEY;NOT NULL"`
	SatelliteID   int `gorm:"UNIQUE_INDEX;NOT NULL"`
	Status        SatelliteStatusID
	LastError     string    `gorm:"NOT NULL"`
	LastErrorTime time.Time `gorm:"NOT NULL"`
	LastExport    time.Time `gorm:"NOT NULL"`
	LastSeen      time.Time `gorm:"NOT NULL"`
}

func (SatelliteStatus) TableName() string {
	return "satellite_status"
}

func (s *SatelliteStatus) Save() {
	s.db.Save(s)
}

// OITC is the database handler
type OITC struct {
	mtx sync.RWMutex
	db  *gorm.DB

	singleInstanceSync bool
}

func readConfigurationIni(path string) (string, error) {
	cfg, err := ini.Load(path)
	if err != nil {
		return "", err
	}
	clientSec, err := cfg.GetSection("client")
	if err != nil {
		return "", err
	}
	hostKey, err := clientSec.GetKey("host")
	if err != nil {
		return "", err
	}
	hostname := hostKey.String()
	userKey, err := clientSec.GetKey("user")
	if err != nil {
		return "", err
	}
	username := userKey.String()

	passwordKey, err := clientSec.GetKey("password")
	if err != nil {
		return "", err
	}
	password := passwordKey.String()

	portKey, err := clientSec.GetKey("port")
	if err != nil {
		return "", err
	}
	port, err := portKey.Int()
	if err != nil {
		port = 3306
	}

	databaseKey, err := clientSec.GetKey("database")
	if err != nil {
		return "", err
	}
	database := databaseKey.String()

	if hostname == "localhost" {
		return fmt.Sprintf("%s:%s@unix(%s)/%s?parseTime=True&loc=Local", username, password, "/run/mysqld/mysqld.sock", database), nil
	}
	return fmt.Sprintf("%s:%s@tcp(%s:%d)/%s?parseTime=True&loc=Local", username, password, hostname, port, database), nil
}

// New connects to the database
func New(cfgPath string) (*OITC, error) {
	url, err := readConfigurationIni(cfgPath)
	if err != nil {
		return nil, err
	}
	return NewFromURL("mysql", url)
}

// NewFromURL connects to the database
func NewFromURL(dialect string, url string) (*OITC, error) {
	db, err := gorm.Open(dialect, url)
	if err != nil {
		return nil, err
	}

	oitc := &OITC{
		db: db,
	}

	return oitc, nil
}

// Close closes the db connection
func (oitc *OITC) Close() {
	oitc.db.Close()
}

// FetchSatellites connects to the mysql server to fetch all satellites in db
func (oitc *OITC) FetchSatellites() []*Satellite {
	satellites := make([]*Satellite, 0)
	oitc.db.Find(&satellites)
	return satellites
}

func (oitc *OITC) FetchSatelliteStatus(satelliteID int) *SatelliteStatus {
	status := &SatelliteStatus{}
	oitc.db.Where(&SatelliteStatus{ID: satelliteID}).Attrs(SatelliteStatus{
		SatelliteID:   satelliteID,
		Status:        SatelliteStatusNil,
		LastError:     "",
		LastErrorTime: time.Time{},
		LastSeen:      time.Time{},
		LastExport:    time.Time{},
	}).FirstOrCreate(status)
	status.db = oitc.db
	return status
}
