package transfer

import (
	"context"
	"fmt"
	"log"
	"sync"
	"time"

	"github.com/it-novum/nsta/gearman"
	"github.com/it-novum/nsta/httpproto"
	"github.com/it-novum/nsta/oitcdb"
	"github.com/it-novum/nsta/satstatus"
)

type PushSatellite struct {
	SatelliteOptions

	wg       sync.WaitGroup
	shutdown chan struct{}

	lastExport time.Time
	toSat      gearman.SyncMessageFIFO

	nextTs   time.Time
	lastSeen time.Time
}

// SatelliteID returns the db id of the satellite
func (s *PushSatellite) SatelliteID() int {
	return s.DB.ID
}

// SatelliteName returns the db name of the satellite
func (s *PushSatellite) SatelliteName() string {
	return s.DB.Name
}

// APIKey returns the db signature api key
func (s *PushSatellite) APIKey() string {
	return s.DB.APIKey
}

func (s *PushSatellite) Incoming(msg *gearman.Message) {
	s.ToMaster <- msg
}

func (s *PushSatellite) Outgoing() []*gearman.Message {
	msgs := s.toSat.GetAll(8192)
	now := time.Now()
	// We insert the timesync message on demand, because we don't want to sync a totally outdated timestamp
	if s.DB.UseTimesync && now.After(s.nextTs) {
		s.nextTs = now.Add(time.Minute)
		ts := gearman.TimesyncMessage{
			SatelliteID: s.DB.ID,
			SetTime:     now,
		}
		if data, err := json.Marshal(ts); err != nil {
			log.Println(&Error{Name: s.DB.Name, Message: "could not create timesync message", Err: err})
		} else {
			msgs = append(msgs, &gearman.Message{
				Type:   gearman.MessageTimesync,
				String: string(data),
			})
		}
	}
	return msgs
}

func (s *PushSatellite) LastExport() time.Time {
	return s.lastExport
}

func (s *PushSatellite) SendMessage(msg *gearman.Message) {
	s.toSat.Push(msg)
}

func (s *PushSatellite) Export(_ *httpproto.ExportData) error {
	return nil
}

func (s *PushSatellite) WebStatus(status *satstatus.Status) {
	if !status.LastSeen.IsZero() {
		s.lastSeen = status.LastSeen
	}
	s.StatusTracker.Update <- status
}

// Run is the main loop of this satellite
func (s *PushSatellite) run(ctx context.Context) {
	timeoutCheck := 300 * time.Second
	timeoutDB := time.Duration(int64(s.DB.Timeout)) * time.Second
	if timeoutDB < 240*time.Second {
		timeoutCheck = 60 * time.Second
	} else if timeoutDB < 3600*time.Second {
		timeoutCheck = timeoutDB / 4
	} else {
		timeoutCheck = 3600 * time.Second
	}

	ts := time.NewTicker(timeoutCheck)

	defer func() {
		log.Println(&Error{Name: s.DB.Name, Message: "shutting down"})
		s.RegisterSatelliteIO.UnregisterSatelliteIO(s.DB.ID)
		ts.Stop()
		log.Println(&Error{Name: s.DB.Name, Message: "close message fifo"})
		s.toSat.Close()
		log.Println(&Error{Name: s.DB.Name, Message: "shutdown complete"})
		s.wg.Done()
	}()
	for {
		select {
		case <-ctx.Done():
			return
		case _, more := <-s.shutdown:
			if !more {
				return
			}
		case <-ts.C:
			absTimeout := s.lastSeen.Add(timeoutDB)
			if !absTimeout.After(time.Now()) {
				log.Println(&Error{Name: s.DB.Name, Message: fmt.Sprintf("Last Seen: %s, Timeout: %s", s.lastSeen.Format(time.RFC3339), absTimeout.Format(time.RFC3339))})
				s.StatusTracker.Update <- &satstatus.Status{
					SatelliteID: s.DB.ID,
					Status:      oitcdb.SatelliteStatusFailed,
					Err:         &Error{Name: s.DB.Name, Message: "timeout reached"},
				}
			}
		}
	}
}

// Close stops the satellite
func (s *PushSatellite) Close() {
	close(s.shutdown)

	s.wg.Wait()
}

// Compare checks if the satellite matches the given db satellite data
func (s *PushSatellite) Compare(dbSat *oitcdb.Satellite) bool {
	return s.DB.Compare(dbSat)
}

func (s *PushSatellite) Reload(_ context.Context) {
	if !s.DB.NSTASyncInstance {
		log.Println(&Error{Name: s.DB.Name, Message: "skip export"})
		return
	}
	s.lastExport = time.Now()
}

func (s *PushSatellite) Start(parent context.Context) error {
	ctx, cancel := context.WithCancel(parent)

	s.shutdown = make(chan struct{})
	s.lastExport = time.Now()

	s.RegisterSatelliteIO.RegisterSatelliteIO(s)
	s.toSat.Start()
	s.wg.Add(1)
	go func() {
		defer s.wg.Done()
		defer cancel()

		s.run(ctx)
	}()

	return nil
}
