package transfer

import (
	"context"
	"fmt"
	"log"
	"path/filepath"
	"sync"
	"time"

	"github.com/it-novum/nsta/checkmk"
	"github.com/it-novum/nsta/gearman"
	"github.com/it-novum/nsta/nstautil"
	"github.com/it-novum/nsta/oitcdb"
	"github.com/it-novum/nsta/satstatus"
	"github.com/it-novum/nsta/sshwrapper"
)

// SSHSatellite manages the connection and message exchange to a satellite via ssh
type SSHSatellite struct {
	// Configuration

	SatelliteOptions

	// Thread Sync

	wg       sync.WaitGroup
	reload   chan chan struct{}
	shutdown chan struct{}

	// SSH + Gearman Connection

	satGearman    *gearman.Client
	ssh           *sshwrapper.Conn
	portPublisher <-chan int

	// State and internal Config

	nagiosConfigPath string
	lastMainExport   time.Time
	toSat            gearman.SyncMessageFIFO

	// Checkmk Commands

	cmk        checkmk.Checkmk
	cmdBackend *sshwrapper.SSHCommandBackend
}

func (s *SSHSatellite) SatelliteID() int {
	return s.DB.ID
}

func (s *SSHSatellite) SatelliteName() string {
	return s.DB.Name
}

// Compare checks if the satellite matches the database config
func (s *SSHSatellite) Compare(dbSat *oitcdb.Satellite) bool {
	return s.DB.Compare(dbSat)
}

// SendMessage to the satellite
func (s *SSHSatellite) SendMessage(msg *gearman.Message) {
	s.toSat.Push(msg)
}

func (s *SSHSatellite) resetPort(ctx context.Context, port int) {
	status := &satstatus.Status{
		SatelliteID: s.DB.ID,
		Status:      oitcdb.SatelliteStatusFailed,
	}
	s.StatusTracker.Update <- status
	if s.satGearman != nil {
		s.satGearman.Stop()
		s.satGearman = nil
	}
	if port != 0 {
		s.satGearman = &gearman.Client{
			ClientType:     gearman.Satellite,
			ClientName:     s.DB.Name,
			Address:        fmt.Sprintf("localhost:%d", port),
			MessageReceive: s.ToMaster,
		}
		s.satGearman.Start(ctx)
		status.Status = oitcdb.SatelliteStatusOK
		status.LastSeen = time.Now()
	}
}

func (s *SSHSatellite) recheckStatus() {
	status := &satstatus.Status{
		SatelliteID: s.DB.ID,
		Status:      oitcdb.SatelliteStatusFailed,
	}
	if s.satGearman != nil && s.satGearman.IsRunning() {
		status.Status = oitcdb.SatelliteStatusOK
		status.LastSeen = time.Now()
	}
	s.StatusTracker.Update <- status
}

func (s *SSHSatellite) export(parent context.Context) {
	ctx, cancel := context.WithTimeout(parent, time.Second*time.Duration(s.SSHReloadTimeoutSec))
	defer cancel()

	if !s.DB.NSTASyncInstance {
		log.Println(&Error{Name: s.DB.Name, Message: "skip export"})
		return
	}

	s.lastMainExport = time.Now()
	status := &satstatus.Status{
		SatelliteID: s.DB.ID,
	}

	defer func() {
		s.StatusTracker.Update <- status
	}()
	log.Println(&Error{Name: s.DB.Name, Message: "copy configuration"})
	tarData := nstautil.CreateExportTar(s.NagiosPath, s.DB.ID, false)
	if tarData != "" {
		if _, err := s.cmdBackend.RunRemoteCommand(ctx, []string{"tar", "xz", "-C", s.nagiosConfigPath}, map[string]string{}, tarData); err != nil {
			status.Err = &Error{Name: s.DB.Name, Message: "export failed", Err: err}
			fmt.Println(status.Err)
			return
		}
		log.Println(&Error{Name: s.DB.Name, Message: "restart monitoring"})
		if _, err := s.cmdBackend.RunRemoteCommand(ctx, []string{"sudo", "/opt/openitc/nagios/bin/restart-monitoring.sh"}, map[string]string{}, ""); err != nil {
			status.Err = &Error{Name: s.DB.Name, Message: "export failed", Err: err}
			fmt.Println(status.Err)
			return
		}
		status.LastExport = time.Now()
		status.LastSeen = time.Now()
		log.Println(&Error{Name: s.DB.Name, Message: "export successful"})
	} else {
		status.Err = &Error{Name: s.DB.Name, Message: "could not create export archive"}
		log.Println(status.Err)
	}
}

// Run is the background go worker
func (s *SSHSatellite) run(ctx context.Context) {
	statusT := time.NewTicker(time.Second * 10)
	noNotify := make(chan struct{})
	var timesyncChan <-chan time.Time
	var timesyncTicker *time.Ticker

	fakeTimesync := make(chan time.Time)
	if s.DB.UseTimesync {
		timesyncTicker = time.NewTicker(time.Minute)
		defer timesyncTicker.Stop()
		timesyncChan = timesyncTicker.C
	} else {
		timesyncChan = fakeTimesync
	}

	defer func() {
		log.Println(&Error{Name: s.DB.Name, Message: "shutting down"})
		statusT.Stop()
		if s.satGearman != nil {
			log.Println(&Error{Name: s.DB.Name, Message: "close gearman connection"})
			s.satGearman.Stop()
			s.satGearman = nil
		}
		log.Println(&Error{Name: s.DB.Name, Message: "stop ssh connection"})
		s.ssh.Stop()
		log.Println(&Error{Name: s.DB.Name, Message: "close message fifo"})
		s.toSat.Close()
		log.Println(&Error{Name: s.DB.Name, Message: "shutdown complete"})
		s.wg.Done()
	}()

	for {
		timesync := timesyncChan
		notify := s.toSat.Notify()
		if s.satGearman == nil {
			timesync = fakeTimesync
			notify = noNotify
		}
		select {
		case <-ctx.Done():
			return
		case _, more := <-s.shutdown:
			if !more {
				return
			}
		case reload := <-s.reload:
			s.export(ctx)
			reload <- struct{}{}
		case <-statusT.C:
			s.recheckStatus()
		case port := <-s.portPublisher:
			s.resetPort(ctx, port)
		case <-notify:
			msg := s.toSat.Get()
			if msg != nil {
				if msg.Type == gearman.MessageCheckmk {
					s.wg.Add(1)
					go func() {
						defer s.wg.Done()
						timeout := time.NewTimer(time.Second * 10)
						defer timeout.Stop()

						response, err := s.cmk.ProcessMessage(s.cmdBackend, msg).Serialize()
						if err != nil {
							log.Println(&Error{Name: s.DB.Name, Message: "could no serialize checkmk answer:", Err: err})
						}
						select {
						case s.ToMaster <- &gearman.Message{
							Type:   gearman.MessageCheckmk,
							String: response,
						}:
						case <-timeout.C:
							log.Println(&Error{Name: s.DB.Name, Message: "could not send checkmk response to gearman: timeout"})
						case <-ctx.Done():
							log.Println(&Error{Name: s.DB.Name, Message: "could not send checkmk response to gearman", Err: ctx.Err()})
						}
					}()
				} else {
					s.satGearman.SendMessage(msg)
				}
			}
		case <-timesync:
			s.wg.Add(1)
			go func() {
				defer s.wg.Done()

				tsCtx, tsCancel := context.WithTimeout(ctx, time.Second*60)
				defer tsCancel()

				if _, err := s.cmdBackend.RunRemoteCommand(tsCtx, []string{"sudo", "date", "-s", fmt.Sprint("@", time.Now().Unix())}, map[string]string{}, ""); err != nil {
					log.Println(err)
				} else {
					log.Println(&Error{Name: s.DB.Name, Message: "time update was successful"})
				}
			}()
		}
	}
}

// Close stops the satellite connection
func (s *SSHSatellite) Close() {
	close(s.shutdown)
	s.wg.Wait()
}

// Reload implements a process reload signal handler
func (s *SSHSatellite) Reload(ctx context.Context) {
	reload := make(chan struct{})

	select {
	case <-ctx.Done():
		return
	case s.reload <- reload:
	}

	select {
	case <-ctx.Done():
	case <-reload:
	}

}

func (s *SSHSatellite) Start(parent context.Context) error {
	s.ssh = &sshwrapper.Conn{
		ConnOpts:   s.SSHConnOpts,
		ClientName: s.DB.Name,
	}

	ctx, cancel := context.WithCancel(parent)
	portPub, err := s.ssh.Start(ctx)
	if err != nil {
		cancel()
		return err
	}
	s.portPublisher = portPub

	s.nagiosConfigPath = filepath.Join(s.NagiosPath, "etc")
	s.lastMainExport = time.Now()
	s.cmdBackend = &sshwrapper.SSHCommandBackend{
		Timeout:    time.Second * 120,
		ConnOpts:   s.SSHConnOpts,
		ClientName: s.DB.Name,
	}
	s.cmk.SatelliteName = s.DB.Name
	s.shutdown = make(chan struct{})
	s.reload = make(chan chan struct{})

	s.toSat.Start()
	s.wg.Add(1)
	go func() {
		defer s.wg.Done()
		defer cancel()
		s.run(ctx)
	}()

	return nil
}
